#include "config.h"

#include <stdio.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <time.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#include <ctype.h>

#include "error.h"
#include "fd.h"
#include "io.h"
#include "display.h"
#include "args.h"

struct _options_list {
	char *short_option1;
	char *short_option2;
	char *long_option1;
	char *long_option2;
	char *arg_description;
	char *description;
	int (*func)(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
};

typedef struct _options_list options_list;

int parse_infile(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_outfile(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_size(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_buffer_size(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_interval(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_timeout(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_kilo(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_width(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_minus_one(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_twiddle(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_count(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_throughput(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_time(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_percent(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_bar(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_summary(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_displays(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_help(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);
int parse_version(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num);

options_list options[] = {
	{
		"-if", 0, "--in-file", 0, "<file>",
		"Read input from <file>.  Default: stdin",
		parse_infile
	},
	{
		"-of", 0, "--out-file", 0, "<file>",
		"Write output to <file>.  Defualt: stdout",
		parse_outfile
	},
	{
		"-s", 0, "--size", 0, "<size>",
		"Expect an input stream of <size> bytes.",
		parse_size
	},
	{
		"-bs", 0, "--buffer-size", 0, "<size>",
		"Allocate an I/O buffer of <size> bytes.  Default: 1024",
		parse_buffer_size
	},
	{
		"-i", 0, "--interval", 0, "<secs>",
		"Update the display every <secs> seconds.  Default: 1",
		parse_interval
	},
	{
		"-t", 0, "--timeout", 0, "<usecs>",
		"Wait <usecs> microseconds for a change in I/O.  Default: 250000",
		parse_timeout
	},
	{
		"-k", 0, "--kilo", 0, "1000|1024",
		"Use 1000 or 1024 as the size of a kilobyte.  Default: 1024",
		parse_kilo
	},
	{
		"-sw", 0, "--screen-width", 0, "<width>",
		"Assume a screen width of <width> characters.  Default: 80",
		parse_width
	},
	{
		"-sw-1", 0, "--screen-width-minus-one", 0, 0,
		"Use one character less than the maximum screen width.  Default: off",
		parse_minus_one
	},
	{
		"-dtw", "-ntw", "--display-twiddle", "--no-twiddle", 0,
		"Turn on/off the twiddle in the display.  Default: on",
		parse_twiddle
	},
	{
		"-dc", "-nc", "--display-count", "--no-count", 0,
		"Turn on/off the data count in the display.  Default: on",
		parse_count
	},
	{
		"-dth", "-nth", "--display-throughput", "--no-throughput", 0,
		"Turn on/off the data throughput in the display.  Default: on",
		parse_throughput
	},
	{
		"-dt", "-nt", "--display-time", "--no-time", 0,
		"Turn on/off the time in the display.  Default: on",
		parse_time
	},
	{
		"-dp", "-np", "--display-percent", "--no-percent", 0,
		"Turn on/off the percent complete in the display.  Default: on",
		parse_percent
	},
	{
		"-db", "-nb", "--display-bar", "--no-bar", 0,
		"Turn on/off the progress bar in the display.  Default: on",
		parse_bar
	},
	{
		"-ds", "-ns", "--display-summary", "--no-summary", 0,
		"Turn on/off the summary information when complete.  Default: on",
		parse_summary
	},
	{
		"-da", "-dn", "--display-all", "--display-none", 0,
		"Turn on/off all displays.  Default: all on",
		parse_displays
	},
	{
		"-h", 0, "--help", 0, 0,
		"Display this help text and exit.",
		parse_help
	},
	{
		"-v", 0, "--version", 0, 0,
		"Display version and exit.",
		parse_version
	},
	{ 0, 0, 0, 0, 0, 0 }
};

void version(FILE *out)
{
	fprintf(out, "%s version %s\n", PACKAGE, VERSION);
}

void help(FILE *out)
{
	int o = 0;
	int screen_used = 0;
	char option_buffer[256];

	fprintf(out, "Usage:\n");
	fprintf(out, "\n");

	/*
	 * Print out short option usage:
	 */
	fprintf(out, "   bar ");
	screen_used = 7;
	for (o = 0; options[o].func != 0; o++) {
		char *short_option1 = options[o].short_option1;
		char *short_option2 = options[o].short_option2;
		char *arg_description = options[o].arg_description;

		sprintf(option_buffer, "[ %s%s%s%s%s ]",
			short_option1,
			(short_option2 != 0) ? "|" : "",
			(short_option2 != 0) ? short_option2 : "",
			(arg_description != 0) ? " " : "",
			(arg_description != 0) ? arg_description : ""
			);
		if (d.screen_width - screen_used < strlen(option_buffer)) {
			fprintf(out, "\n       ");
			screen_used = 7;
		}
		fprintf(out, option_buffer);
		screen_used += strlen(option_buffer);
	}

	fprintf(out, "\n");
	fprintf(out, "\n");

	/*
	 * Print out long option usage:
	 */
	fprintf(out, "   bar ");
	screen_used = 7;
	for (o = 0; options[o].func != 0; o++) {
		char *long_option1 = options[o].long_option1;
		char *long_option2 = options[o].long_option2;
		char *arg_description = options[o].arg_description;

		sprintf(option_buffer, "[ %s%s%s%s%s ]",
			long_option1,
			(long_option2 != 0) ? "|" : "",
			(long_option2 != 0) ? long_option2 : "",
			(arg_description != 0) ? " " : "",
			(arg_description != 0) ? arg_description : ""
			);
		if (d.screen_width - screen_used < strlen(option_buffer)) {
			fprintf(out, "\n       ");
			screen_used = 7;
		}
		fprintf(out, option_buffer);
		screen_used += strlen(option_buffer);
	}

	fprintf(out, "\n");
	fprintf(out, "\n");

	/*
	 * Print out help text
	 */
	for (o = 0; options[o].func != 0; o++) {
		char *short_option1 = options[o].short_option1;
		char *short_option2 = options[o].short_option2;
		char *long_option1 = options[o].long_option1;
		char *long_option2 = options[o].long_option2;
		char *arg_description = options[o].arg_description;
		char *description = options[o].description;

		fprintf(out, "   %s %s\n",
			short_option1,
			(arg_description != 0) ? arg_description : ""
			);
		if (short_option2 != 0) {
			fprintf(out, "   %s %s\n",
				short_option2,
				(arg_description != 0) ? arg_description : ""
				);
		}
		fprintf(out, "   %s %s\n",
			long_option1,
			(arg_description != 0) ? arg_description : ""
			);
		if (long_option2 != 0) {
			fprintf(out, "   %s %s\n",
				long_option2,
				(arg_description != 0) ? arg_description : ""
				);
		}
		fprintf(out, "\n");
		fprintf(out, "     %s\n", description);
		fprintf(out, "\n");
	}
}

int isOpt(char *s)
{
	int o = 0;

	while (options[o].func != 0) {
		if (strcmp(s, options[o].short_option1) == 0)
			return(o);
		if ((options[o].short_option2 != 0) 
			&& (strcmp(s, options[o].short_option2) == 0))
			return(o);
		if (strcmp(s, options[o].long_option1) == 0)
			return(o);
		if ((options[o].long_option2 != 0) 
			&& (strcmp(s, options[o].long_option2) == 0))
			return(o);
		o++;
	}
	return(-1);
}

int safe_add(uint64 *n, uint64 a)
{
	uint64 t = *n;

	if (MAX_UINT64 - t < a)
		return(1);
	t += a;
	*n = t;
	return(0);
}

int safe_mul(uint64 *n, uint64 x)
{
	uint64 a1, a2;
	uint64 t = 0;
	uint64 multiplier;
	
	if (x < *n) {
		a1 = *n;
		a2 = x;
	}
	else {
		a2 = *n;
		a1 = x;
	}

	multiplier = 1;
	multiplier *= 1000;
	multiplier *= 1000;
	multiplier *= 1000;
	multiplier *= 1000;
	multiplier *= 1000;
	multiplier *= 1000;
	while (a2 > 0) {
		while (multiplier > a2)
			multiplier /= 10;
		if (safe_add(&t, a1*multiplier) != 0)
			return(1);
		a2 -= multiplier;
	}
	*n = t;
	return(0);
}

int parse_num(FILE *ferr, char *s, uint64 *n)
{
	char *ptr_decimal = 0;
	char *ptr_unit = 0;
	char *ptr_start = 0;
	char *ptr_end = 0;
	char *ptr = 0;
	uint64 w = 0;
	uint64 f = 0;
	uint64 unit_multiplier = 1;
	uint64 multiplier = 0;
	uint64 tmp = 0;
	uint64 new_n = 0;

	*n = 0;

	if (strlen(s) == 0)
		return(0);

	ptr_decimal = s;
	while ((*ptr_decimal != '\0') && (*ptr_decimal != '.'))
		ptr_decimal++;
	
	ptr_unit = s;
	while ((*ptr_unit != '\0') && (isdigit((int)*ptr_unit) || (*ptr_unit == '.')))
		ptr_unit++;
	if (*(ptr_unit+1) != '\0') {
		print_error(ferr, "Cannot parse number: %s", s);
		print_esup(ferr, "Unit multiplier parse error at: %s", ptr_unit);
		print_esup(ferr, "Unit multiplier should be the last character");
		return(1);
	}

	switch (toupper(*ptr_unit)) {
		case '\0':
			break;
		case 'K':
			unit_multiplier *= (uint64)d.k;
			break;
		case 'M':
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			break;
		case 'G':
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			break;
		case 'T':
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			break;
		case 'P':
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			break;
		case 'E':
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			unit_multiplier *= (uint64)d.k;
			break;
		default:
			print_error(ferr, "Cannot parse number: %s", s);
			print_esup(ferr, "Invalid unit multiplier: '%c'", *ptr_unit);
			print_esup(ferr, "Expected: 'k', 'm', 'g', 't', 'p', or 'e'");
			return(1);
			break;
	}

	ptr_start = s;
	if (*ptr_decimal == '.')
		ptr_end = ptr_decimal;
	else
		ptr_end = ptr_unit;
	
	multiplier = 1;
	if (ptr_start != ptr_end) {
		for (ptr = ptr_start+1; ptr != ptr_end; ptr++) {
			if (safe_mul(&multiplier, 10) != 0) {
				print_error(ferr, "Whole number precision error at: %s", ptr);
				print_esup(ferr, "Whole number portion too large");
				return(1);
			}
		}
	}
	for (ptr = ptr_start; ptr != ptr_end; ptr++) {
		tmp = (uint64)((*ptr) - '0');
		if (tmp != 0) {
			if (safe_mul(&tmp, multiplier) != 0) {
				print_error(ferr, "Multiplication overflow error");
				print_esup(ferr, "Could not parse number at: %s", ptr);
				return(1);
			}
			if (safe_add(&w, tmp) != 0) {
				print_error(ferr, "Addition overflow error");
				print_esup(ferr, "Could not parse number at: %s", ptr);
				return(1);
			}
		}
		multiplier /= 10;
	}
	if (safe_mul(&w, unit_multiplier) != 0) {
		print_error(ferr, "Multiplication overflow error");
		print_esup(ferr, "Number too large: %s", s);
		return(1);
	}

	if (*ptr_decimal == '.') {
		ptr_start = ptr_decimal+1;
		ptr_end = ptr_unit;
		multiplier = 1;
		for (ptr = ptr_start; ptr != ptr_end; ptr++) {
			tmp = (*ptr) - '0';
			if (tmp != 0) {
				if (safe_mul(&tmp, unit_multiplier) != 0) {
					print_error(ferr, "Multiplication overflow error");
					print_esup(ferr, "Could not parse fraction at: %s", ptr);
					return(1);
				}
				tmp /= multiplier;
				tmp += 5;
				tmp /= 10;
				if (safe_add(&f, tmp) != 0) {
					print_error(ferr, "Addition overflow error");
					print_esup(ferr, "Could not parse fraction at: %s", ptr);
				}
			}
			if (safe_mul(&multiplier, 10) != 0) {
				print_error(ferr, "Multiplication overflow error");
				print_esup(ferr, "Could not parse fraction at: %s", ptr);
				return(1);
			}
			if (multiplier > unit_multiplier)
				break;
		}
	}

	new_n = w;
	if (safe_add(&new_n, f) != 0) {
		print_error(ferr, "Addition overflow error");
		print_esup(ferr, "Number too large: %s", s);
	}
	*n = new_n;

	return(0);
}

int parse_infile(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	if ((*arg_num)+1 == argc) {
		print_error(ferr, "Missing filename after %s", argv[(*arg_num)]);
		return(1);
	}
	if (strcmp(argv[(*arg_num)+1], "-") != 0) {
		io.in = open(argv[(*arg_num)+1], O_RDONLY
#ifdef O_LARGEFILE
		|O_LARGEFILE
#endif
		);
		if (io.in < 0) {
			print_error(ferr, "Cannot open file for reading: %s",
				argv[(*arg_num)+1]);
			return(1);
		}
		if (fdIsFile(io.in) && fdFileSize(io.in, &io.total_size) == 0) {
			io.total_size_known = 1;
		}
	}
	(*arg_num)++;
	return(0);
}

int parse_outfile(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	if ((*arg_num)+1 == argc) {
		print_error(ferr, "Missing filename after %s", argv[(*arg_num)]);
		return(1);
	}

	if (strcmp(argv[(*arg_num)+1], "-") != 0) {
		io.out = open(argv[(*arg_num)+1], O_WRONLY|O_CREAT
#ifdef O_LARGEFILE
		|O_LARGEFILE
#endif
		, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH);
		if (io.out < 0) {
			print_error(ferr, "Cannot open file for writing: %s",
				argv[(*arg_num)+1]);
			return(1);
		}
	}
	(*arg_num)++;
	return(0);
}

int parse_size(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	uint64 n = 0;

	if ((*arg_num)+1 == argc) {
		print_error(ferr, "Missing size after %s", argv[(*arg_num)]);
		return(1);
	}
	if (parse_num(ferr, argv[(*arg_num)+1], &n) != 0) {
		print_error(ferr, "Type mismatch or number too large: %s",
			argv[(*arg_num)+1]);
		print_esup(ferr, "Ignoring given size");
		(*arg_num)++;
		return(0);
	}
	io.total_size = n;
	io.total_size_known = 1;
	(*arg_num)++;
	return(0);
}

int parse_buffer_size(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	uint64 n = 0;

	if ((*arg_num)+1 == argc) {
		print_error(ferr, "Missing size after %s", argv[(*arg_num)]);
		return(1);
	}
	if (parse_num(ferr, argv[(*arg_num)+1], &n) != 0) {
		print_error(ferr, "Type mismatch or number too large: %s",
			argv[(*arg_num)+1]);
		print_esup(ferr, "Ignoring given buffer size");
		(*arg_num)++;
		return(0);
	}
	io.buffer_size = (size_t)n;
	(*arg_num)++;
	return(0);
}

int parse_interval(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	int n = 0;
	char check[4096] = { 0 };

	if ((*arg_num)+1 == argc) {
		print_error(ferr, "Missing seconds after %s", argv[(*arg_num)]);
		return(1);
	}
	sscanf(argv[(*arg_num)+1], "%d", &n);
	sprintf(check, "%d", n);
	if (strcmp(argv[(*arg_num)+1], check) != 0) {
		print_error(ferr, "Type mismatch or number too large: %s",
			argv[(*arg_num)+1]);
		print_esup(ferr, "Ignoring given interval");
		(*arg_num)++;
		return(0);
	}
	if ((n < 1) || (n > 60*60*24)) {
		print_error(ferr, "Invalid display interval: %s",
			argv[(*arg_num)+1]);
		return(1);
	}
	d.display_interval = n;
	(*arg_num)++;
	return(0);
}

int parse_timeout(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	unsigned long n = 0;
	char check[4096] = { 0 };

	if ((*arg_num)+1 == argc) {
		print_error(ferr, "Missing microseconds after %s", argv[(*arg_num)]);
		return(1);
	}
	sscanf(argv[(*arg_num)+1], "%lu", &n);
	sprintf(check, "%lu", n);
	if (strcmp(argv[(*arg_num)+1], check) != 0) {
		print_error(ferr, "Type mismatch or number too large: %s",
			argv[(*arg_num)+1]);
		print_esup(ferr, "Ignoring given timeout");
		(*arg_num)++;
		return(0);
	}
	if (n > 999999) {
		print_error(ferr, "Type mismatch or number too large: %s",
			argv[(*arg_num)+1]);
		print_esup(ferr, "Ignoring given timeout");
		(*arg_num)++;
		return(0);
	}
	io.timeout = (uint32)n;
	(*arg_num)++;
	return(0);
}

int parse_kilo(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	if ((*arg_num)+1 == argc) {
		print_error(ferr, "Missing option after %s", argv[(*arg_num)]);
		return(1);
	}
	if (strcmp(argv[(*arg_num)+1], "1000") == 0) {
		d.k = 1000;
	}
	else if (strcmp(argv[(*arg_num)+1], "1024") == 0) {
		d.k = 1024;
	}
	else {
		print_error(ferr, "Invalid option for %s: %s",
			argv[(*arg_num)], argv[(*arg_num)+1]);
		return(1);
	}
	(*arg_num)++;
	return(0);
}

int parse_width(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	int n = 0;

	if ((*arg_num)+1 == argc) {
		print_error(ferr, "Missing width after %s", argv[(*arg_num)]);
		return(1);
	}
	sscanf(argv[(*arg_num)+1], "%d", &n);
	if (n < 1) {
		print_error(ferr, "Invalid screen width: %s", argv[(*arg_num)+1]);
		return(1);
	}
	d.screen_width = n;
	d.manual_width = 1;
	(*arg_num)++;
	return(0);
}

int parse_switch(char *arg, 
	char *so1, char *so2, char *lo1, char *lo2, int *sw)
{
	if ((so1 != 0) && (strcmp(arg, so1) == 0))
		(*sw) = 1;
	if ((lo1 != 0) && (strcmp(arg, lo1) == 0))
		(*sw) = 1;
	if ((so2 != 0) && (strcmp(arg, so2) == 0))
		(*sw) = 0;
	if ((lo2 != 0) && (strcmp(arg, lo2) == 0))
		(*sw) = 0;
	return(0);
}

int parse_switch_all(char *arg, 
	char *so1, char *so2, char *lo1, char *lo2)
{
	if ((so1 != 0) && (strcmp(arg, so1) == 0)) {
		d.display_twiddle = 1;
		d.display_datacount = 1;
		d.display_throughput = 1;
		d.display_time = 1;
		d.display_percent = 1;
		d.display_bar = 1;
		d.display_summary = 1;
	}
	if ((lo1 != 0) && (strcmp(arg, lo1) == 0)) {
		d.display_twiddle = 1;
		d.display_datacount = 1;
		d.display_throughput = 1;
		d.display_time = 1;
		d.display_percent = 1;
		d.display_bar = 1;
		d.display_summary = 1;
	}
	if ((so2 != 0) && (strcmp(arg, so2) == 0)) {
		d.display_twiddle = 0;
		d.display_datacount = 0;
		d.display_throughput = 0;
		d.display_time = 0;
		d.display_percent = 0;
		d.display_bar = 0;
		d.display_summary = 0;
	}
	if ((lo2 != 0) && (strcmp(arg, lo2) == 0)) {
		d.display_twiddle = 0;
		d.display_datacount = 0;
		d.display_throughput = 0;
		d.display_time = 0;
		d.display_percent = 0;
		d.display_bar = 0;
		d.display_summary = 0;
	}
	return(0);
}

int parse_minus_one(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	int r;
	
	r = parse_switch(
		argv[(*arg_num)],
		options[option_num].short_option1,
		options[option_num].short_option2,
		options[option_num].long_option1,
		options[option_num].long_option2,
		&d.screen_width_minus_one
		);
	return(r);
}

int parse_twiddle(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	int r;
	
	r = parse_switch(
		argv[(*arg_num)],
		options[option_num].short_option1,
		options[option_num].short_option2,
		options[option_num].long_option1,
		options[option_num].long_option2,
		&d.display_twiddle
		);
	return(r);
}

int parse_count(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	int r;
	
	r = parse_switch(
		argv[(*arg_num)],
		options[option_num].short_option1,
		options[option_num].short_option2,
		options[option_num].long_option1,
		options[option_num].long_option2,
		&d.display_datacount
		);
	return(r);
}

int parse_throughput(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	int r;
	
	r = parse_switch(
		argv[(*arg_num)],
		options[option_num].short_option1,
		options[option_num].short_option2,
		options[option_num].long_option1,
		options[option_num].long_option2,
		&d.display_throughput
		);
	return(r);
}

int parse_time(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	int r;
	
	r = parse_switch(
		argv[(*arg_num)],
		options[option_num].short_option1,
		options[option_num].short_option2,
		options[option_num].long_option1,
		options[option_num].long_option2,
		&d.display_time
		);
	return(r);
}

int parse_percent(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	int r;
	
	r = parse_switch(
		argv[(*arg_num)],
		options[option_num].short_option1,
		options[option_num].short_option2,
		options[option_num].long_option1,
		options[option_num].long_option2,
		&d.display_percent
		);
	return(r);
}

int parse_bar(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	int r;
	
	r = parse_switch(
		argv[(*arg_num)],
		options[option_num].short_option1,
		options[option_num].short_option2,
		options[option_num].long_option1,
		options[option_num].long_option2,
		&d.display_bar
		);
	return(r);
}

int parse_summary(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	int r;
	
	r = parse_switch(
		argv[(*arg_num)],
		options[option_num].short_option1,
		options[option_num].short_option2,
		options[option_num].long_option1,
		options[option_num].long_option2,
		&d.display_summary
		);
	return(r);
}

int parse_displays(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	int r;
	
	r = parse_switch_all(
		argv[(*arg_num)],
		options[option_num].short_option1,
		options[option_num].short_option2,
		options[option_num].long_option1,
		options[option_num].long_option2
		);
	return(r);
}

int parse_help(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	help(ferr);
	return(1);
}

int parse_version(FILE *ferr, int argc, char *argv[], int *arg_num, int option_num)
{
	version(ferr);
	return(1);
}

int parse_args(FILE *ferr, int argc, char *argv[])
{
	int c = 0;

	for (c = 1; c < argc; c++) {
		int o = 0;

		o = isOpt(argv[c]);
		if (o == -1) {
			print_error(ferr, "Unknown command line option: %s", argv[c]);
			return(1);
		}
		if (options[o].func == parse_kilo) {
			if (options[o].func(ferr, argc, argv, &c, o) != 0)
				return(1);
		}
		else if (options[o].func == parse_help) {
			if (options[o].func(ferr, argc, argv, &c, o) != 0)
				return(1);
		}
		else if (options[o].func == parse_version) {
			if (options[o].func(ferr, argc, argv, &c, o) != 0)
				return(1);
		}
		else if (options[o].arg_description != 0) {
			c++;
		}
	}

	for (c = 1; c < argc; c++) {
		int o = 0;
		o = isOpt(argv[c]);
		if (options[o].func(ferr, argc, argv, &c, o) != 0)
			return(1);
	}
	return(0);
}

